# initializeWormFitting.py
#
# Bryan Daniels
# 4.4.2016
#
# Use SirIsaac's multiple condition fitting to get a single model that
# describes three conditions for the worm startle response as a function
# of laser intensity: "control", "ibuprofen", and "mutant".
#

from SirIsaac import FittingProblemMultipleCondition
from SirIsaac.fitAllParallel import initializeFitAllParallel
import scipy, sys
import scipy.signal
from SirIsaac.outputTag import nextFileNumString
from SirIsaac.simplePickle import save
from pulseInput import *
import explainableVariance as ev

# set tag to command-line argument unless there isn't a valid one
outputDirectory = '.'
if len(sys.argv) < 2:
    prefix = nextFileNumString(outputDirectory)
elif (sys.argv[1].find('&') != -1) or (sys.argv[1].find('>') != -1)         \
    or (sys.argv[1].find('|') != -1):
    prefix = nextFileNumString(outputDirectory)
else:
    prefix = sys.argv[1]


# () Set fitting parameters

complexityStepsize = 1 # increase complexity with steps of size 1
complexityMax = 25 # don't try models with complexity > 25
complexityList = range(0,complexityMax,complexityStepsize) 

# ensGen controls the generation of the initial ensemble of 
# parameter starting points.
totalSteps = 1e3 #10
keepSteps = 10
seeds = (1,1) # use a fixed random seed
ensTemperature = 100. #1000. #100.
ensGen = FittingProblemMultipleCondition.EnsembleGenerator( \
    totalSteps, keepSteps, temperature=ensTemperature, seeds=seeds )

# Parameters that control when local fitting stops.
avegtol = 1e-2
maxiter = 100
verbose = False
stopFittingN = 5 #6 #3

# Order of adding connections
connectionOrder = 'node'
typeOrder = 'last'
connectionOrderSeed = None

# priorSigma controls the width of priors on all parameters
priorSigma = 100.

# If you have pypar installed, you can run on multiple processors
numprocs = 10


# () Set parameters controlling data to use

# as of 6.26.2017, constStd is old estimate of experimental uncertainty
constStd = 10. # 20
# as of 6.26.2017, useVariableStd is the new way to estimate experimental uncertainty
useVariableStd = True #True # False # effectively False 0034 and prior

# We won't use all the data due to correlation structure
maxTimesPerIndepParam = 4 #8 #16 #6 #3
deltaNumDatapoints = 2 #100

# use to only quickly generate data
#maxTimesPerIndepParam = 1 #  only quickly generate data
#deltaNumDatapoints = 201 # only quickly generate data


# which random times to choose
timeSeed = 124 #1248 for 0042 #(124 used for 0041 and 0034, likely others) #123

# filter out trials that don't produce a large enough response
# (and optionally keep small-response trials for weak laser intensities)
thresholdMaxSpeed = None #10 # pixels/sec # None # *** changed 5.19.2017, 6.26.2017
keepPaused = True # effectively False 0037 and prior
#                   (though only has effect when thresholdMaxSpeed is not None)
keepPausedBelowThreshold = False # effectively False 0036 and prior
laserThreshold = 25.9 # mA (laser intensity threshold used with keepPausedBelowThreshold)
laserFilterRange = None #[20.,30.] # None # mA (filter out trials in this range of laser int.)
                             # effectively None 0037 and prior

# include as input whether the trial remains paused
pauseStateAsInput = False #True # effectively False 0037 and prior

# optionally use Gaussian smoothed data
smoothed = True # False
# 11.14.2018 set width of smoothing
# (if None and smoothed=True, use loaded pre-smoothed data)
# effectively None (or roughly 0.07448364) 0042 and prior
gaussianSmoothingSigma = None #2. * 0.07448364 #/ 2. # s

# which experimental conditions to use
# 0=control, 1=ibuprofen, 2=mutant
conditionIndices = [0] #[0,1,2]

# in-sample trials to use per condition
numInSampleTrials = None #100 # None

# have laser input follow pulse dynamics
pulse = True # False
pulseType = 'riseAndDecay' # 'pulse'
pulseDuration = 0.1 # sec
pulseDelay = 1. #0. #0.15 # sec
delayOptimizable = False
delayAsInput = True # False
delaySeed = 1234
pulseTau = 0.25 # sec
tauOptimizable = False
equilibrationTime = 10 #10000. #10. # sec
cutAtLaser = True # effectively run with False last for 0021
equalWeightPreLaser = True # effectively run with False last for 0028
newCutAtLaser = True # effectively run with False last for 0029


# () set up filename for output
fileNumString = prefix
print "initializeWormFitting: Output files will start with",fileNumString
configString = '_fitProb_varying_numInputs'                                 \
   +'_wormData'                                                             \
   +'_CTSN'                                                                 \
   +'_withEnsembleT'+str(int(ensTemperature))                               \
   +'_steps'+str(totalSteps)+'_'+str(keepSteps)                             \
   +'_maxiter'+str(maxiter)                                                 \
   +'_avegtol'+str(avegtol)                                                 \
   +'_timeSeed'+str(timeSeed)                                               \
   +'_thresholdMaxSpeed'+str(thresholdMaxSpeed)
saveFilename = fileNumString+configString+'.dat'

# 11.14.2018 for smoothing data here instead of using pre-smoothed data
# (use gaussianSmoothingSigma above)
def gaussianSmooth(noisyVals,sigma):
    """
    Sigma is measured in units of timesteps.
    """
    window = scipy.signal.gaussian(M=len(noisyVals),std=sigma)
    return scipy.signal.convolve(noisyVals,window,mode='same') / scipy.sum(window)

def sloppyCellWormData(matlabData,thresholdMaxSpeed=None,smoothed=False,
                       variableDelay=False,delaySeed=1234,
                       equilibrationTime=0.,cutAtLaser=True,newCutAtLaser=True,
                       equalWeightPreLaser=True,preLaserSeed=12345,
                       preLaserJitter=0.0001,useVariableStd=False,
                       keepPaused=False,
                       keepPausedBelowThreshold=False,laserThreshold=25.9,
                       pauseStateAsInput=False,laserFilterRange=None,
                       gaussianSmoothingSigma=None):
    """
    thresholdMaxSpeed (None)    : If given a value, defines paused trials
                                  as those
                                  that have a maximum smoothed reverse 
                                  speed < thresholdMaxSpeed pixels/sec.
                                  (Typical thresholdMaxSpeed = +10)
    keepPaused (False)          : In conjunction with thresholdMaxSpeed,
                                  filters out paused states by default.
                                  Does nothing if thresholdMaxSpeed is None.
    keepPausedBelowThreshold    : In conjunction with thresholdMaxSpeed
    (False)                       and laserThreshold, set True to keep 
                                  ONLY trials below laser threshold and paused
                                  or above laser threshold and not paused.
                                  Overrides keepPaused setting.
    equilibrationTime (0.)      : Time before data starts.  (Earliest
                                  datapoint has time equal to
                                  equilibrationTime)
    laserFilterRange (None)     : If a length-2 list, filter out trials 
                                  with laser threshold between 
                                  laserFilterRange[0] and laserFilterRange[1].
                                  Works independently of other trial filtering.
    """
    laserIntensities = matlabData['I'][0] # (numTrials)
    # speedsList has shape (numTrials)x(#times)
    speedsList = scipy.transpose(matlabData['speed']) # pixels / sec (100 pixel ~ 1mm)
    smoothedSpeedsList = scipy.transpose(matlabData['fspeed'])
    times = scipy.arange(len(speedsList[0]))*1./60. # sec
    
    # optionally use only trials with large enough reverse speed
    if (thresholdMaxSpeed is None) and pauseStateAsInput: raise Exception
    if (not keepPaused) and (not keepPausedBelowThreshold) and pauseStateAsInput:
        raise Exception
    if thresholdMaxSpeed is not None:
        minTimeThreshold = 1. # sec
        maxTimeThreshold = 200./60. # sec
        thresholdTimeIndices = \
            filter(lambda i: minTimeThreshold <= times[i] <= maxTimeThreshold,
                   range(len(times)))
        maxRevSpeed = scipy.amax(-smoothedSpeedsList[:,thresholdTimeIndices],axis=1)
        pauseStateList = maxRevSpeed < thresholdMaxSpeed
        pauseTrials = filter(lambda i: pauseStateList[i],range(len(laserIntensities)))
        if keepPausedBelowThreshold: # 9.8.2017
            trialIndicesToUse = filter(lambda i:                \
                ( (laserIntensities[i] >= laserThreshold) and   \
                  (i not in pauseTrials) )                      \
                or                                              \
                ( (laserIntensities[i] < laserThreshold) and    \
                  (i in pauseTrials) ),
                range(len(laserIntensities)))
        elif keepPaused:
            trialIndicesToUse = range(len(laserIntensities))
        else: # filter out paused trials (default when thresholdMaxSpeed is set)
            trialIndicesToUse = filter(lambda i: i not in pauseTrials,
                                       range(len(laserIntensities)))
    else:
        trialIndicesToUse = range(len(laserIntensities))
    if laserFilterRange is not None:
        trialIndicesToUse = filter(lambda i: \
                              (laserIntensities[i] < laserFilterRange[0]) or
                              (laserIntensities[i] > laserFilterRange[1]),
                              trialIndicesToUse )
    print "sloppyCellWormData: Using",len(trialIndicesToUse),"total trials (in- and out-of-sample)."

    if gaussianSmoothingSigma is not None:
        gaussianSigmaInTimesteps = gaussianSmoothingSigma * 60 # timesteps
        dataList = scipy.array([ gaussianSmooth(speeds,gaussianSigmaInTimesteps) for speeds in speedsList ])
    elif smoothed:
        dataList = smoothedSpeedsList
    else:
        dataList = speedsList

    # set up inputs
    laserIntensitiesToUse = [ [l] for l in laserIntensities[trialIndicesToUse] ]
    if pauseStateAsInput:
        inputList = [ [laser,pause] for laser,pause             \
                      in zip(laserIntensities[trialIndicesToUse],
                             pauseStateList[trialIndicesToUse]) ]
    else:
        inputList = laserIntensitiesToUse

    # calculate uncertainties
    if useVariableStd:
        # 6.26.2017 estimate "unexplainable variance"
        numbins = 5
        stdData = ev.unexplainableStdev(dataList[trialIndicesToUse],
                                        laserIntensitiesToUse,
                                        numbins)
    else:
        stdData = constStd * scipy.ones_like(dataList[trialIndicesToUse])

    # if using variable delay, choose delay between 0 and 1 sec (0 and 60 frames)
    if variableDelay:
        scipy.random.seed(delaySeed)
        delayList = scipy.random.choice(times[0:60],size=len(trialIndicesToUse))
    else:
        delayList = scipy.zeros(len(trialIndicesToUse))

    data = []
    for speeds,delay,stdSpeeds in zip(dataList[trialIndicesToUse],delayList,stdData):
        
        # chop off early and late times
        if not cutAtLaser: # used last for run 0021
            minTime = 1. - delay #0. #1. # sec
            maxTime = 3.25 + minTime # sec
        else:
            minTime = 0.5  # sec
            maxTime = 3.25 #4.5 #3.25 # sec
            laserStart = 1.# sec
        # avoid using t = 0
        # see https://github.com/EmoryUniversityTheoreticalBiophysics/SirIsaac/issues/5
        timeIndicesToUse = filter(lambda i: minTime < times[i] <= maxTime,
                                  range(len(times)))

        if not cutAtLaser: # used last for run 0021
            if newCutAtLaser: raise Exception
            trialData = { 'speed': \
                dict([( times[i] - minTime + equilibrationTime,
                        (speeds[i],stdSpeeds[i]) ) \
                      for i in timeIndicesToUse]) }
        else: # cut into two pieces, one pre-laser and one post-laser, separated by delay
            trialData = { 'speed': {} }
            
            # () determine delayed time
            if equalWeightPreLaser:
                # include multiple copies of data pre-laser in order to
                # have the same amount of data pre- and post-laser
                scipy.random.seed(preLaserSeed)
                timeIndicesPreLaser = filter(lambda i: times[i] < laserStart,
                                             timeIndicesToUse)
                timeIndicesPostLaser = filter(lambda i: times[i] >= laserStart,
                                              timeIndicesToUse)
                timeIndicesToUsePreLaser = scipy.random.choice(timeIndicesPreLaser,
                                                               len(timeIndicesPostLaser))
                timeIndicesToUse = scipy.concatenate([timeIndicesToUsePreLaser,
                                                      timeIndicesPostLaser])
            for i in timeIndicesToUse:
              t = copy.copy(times[i])
              if t < laserStart: # pre-laser
                if equalWeightPreLaser:
                    # include jitter so duplicate datapoints are not simply overwritten
                    t += preLaserJitter*(scipy.random.random()-0.5)
                if newCutAtLaser: # 2.27.2017
                    # put pre-laser data either just before the laser or
                    # at delay before the laser
                    if scipy.random.choice(2):
                        time = t - minTime #+ equilibrationTime/2.
                    else:
                        time = t - laserStart + equilibrationTime + delay
                else: # old cutAtLaser, used prior to 0029 2.27.2017
                    time = t - minTime
              else: # post-laser
                time = t - laserStart + equilibrationTime + delay
              
              # record speed
              trialData['speed'].update( {time: (speeds[i],stdSpeeds[i])} )
        data.append(trialData)

    # shift delay times by equilibrationTime
    delayList = scipy.array(delayList) + equilibrationTime

    return inputList, delayList, data

# *****************
# for debugging
#limitNumberOfTrials = 10
# *****************

# () Load data
dataDir = '../Data/160325/'
controlData = scipy.io.loadmat(dataDir+'control_oct.mat')
#ibuprofenData = scipy.io.loadmat(dataDir+'ibuprofen.mat')
#mutantData = scipy.io.loadmat(dataDir+'mutant.mat')
#conditionsData = [ [controlData,ibuprofenData,mutantData][i] for i in conditionIndices ]
conditionsData = [ [controlData,controlData,controlData][i] for i in conditionIndices ]
fittingDataMultiple,delayListMultiple,indepParamsListMultiple = [],[],[]
for i,d in enumerate(conditionsData):
    scd = sloppyCellWormData(d,thresholdMaxSpeed=thresholdMaxSpeed,
                             smoothed=smoothed,delaySeed=i+delaySeed,
                             variableDelay=delayAsInput,
                             equilibrationTime=equilibrationTime,
                             cutAtLaser=cutAtLaser,newCutAtLaser=newCutAtLaser,
                             equalWeightPreLaser=equalWeightPreLaser,
                             useVariableStd=useVariableStd,
                             keepPaused=keepPaused,
                             keepPausedBelowThreshold=keepPausedBelowThreshold,
                             laserThreshold=laserThreshold,
                             pauseStateAsInput=pauseStateAsInput,
                             laserFilterRange=laserFilterRange,
                             gaussianSmoothingSigma=gaussianSmoothingSigma)
    indepParamsListMultiple.append(scd[0])#[:limitNumberOfTrials])
    delayListMultiple.append(scd[1])#[:limitNumberOfTrials])
    fittingDataMultiple.append(scd[2])#[:limitNumberOfTrials])
outputNames = ['speed']
if pauseStateAsInput: # 9.26.2017
    indepParamNames = ['laserIntensity','pauseState']
else:
    indepParamNames = ['laserIntensity']


# () Make the full fittingProblem with all the data
kwargs = {  'avegtol': avegtol,
            'maxiter': maxiter,
            'ensGen': ensGen,
            'verbose': verbose,
            'indepParamNames': indepParamNames,
            'indepParamsListMultiple': indepParamsListMultiple,
            'saveFilename': saveFilename,
            'numprocs': numprocs,
            'saveKey': None,
            'stopFittingN': stopFittingN,
            'connectionOrder': connectionOrder,
            'typeOrder': typeOrder,
            'connectionOrderSeed': connectionOrderSeed,
            'outputNames': outputNames,
            'priorSigma': priorSigma,
          }
p = FittingProblemMultipleCondition.CTSNFittingProblemMultipleCondition(\
                                    complexityList,fittingDataMultiple,**kwargs)

# () Optionally change input to dynamic pulse
if pulse:
    for f,delayList in zip(p.fittingProblemList,delayListMultiple):
        pulseInput(f,indepParamNames[0],
                   pulseDuration,dynamicsType=pulseType,
                   delay=pulseDelay,delayOptimizable=delayOptimizable,
                   tau=pulseTau,tauOptimizable=tauOptimizable,
                   delayAsInput=delayAsInput,delayList=delayList)

# () Save full multiple condition fitting problem including all data
save(p,saveFilename[:-4]+'_allData.dat')

# () Save configuration dictionary for reference
configDict = {'timeSeed': timeSeed,
              'thresholdMaxSpeed': thresholdMaxSpeed,
              'smoothed': smoothed,
              'maxTimesPerIndepParam': maxTimesPerIndepParam,
              'constStd': constStd,
              'useVariableStd': useVariableStd,
              'pulse': pulse,
              'pulseType': pulseType,
              'pulseDuration': pulseDuration,
              'pulseDelay': pulseDelay,
              'delayOptimizable': delayOptimizable,
              'delayAsInput': delayAsInput,
              'delayListMultiple': delayListMultiple,
              'delaySeed': delaySeed,
              'pulseTau': pulseTau,
              'equilibrationTime': equilibrationTime,
              'tauOptimizable': tauOptimizable,
              'numprocs': numprocs,
              'priorSigma': priorSigma,
              'kwargs': kwargs,
              'conditionIndices': conditionIndices,
              'numInSampleTrials': numInSampleTrials,
              'cutAtLaser': cutAtLaser,
              'newCutAtLaser': newCutAtLaser,
              'keepPausedBelowThreshold': keepPausedBelowThreshold,
              'keepPaused': keepPaused,
              'laserThreshold': laserThreshold,
              'pauseStateAsInput': pauseStateAsInput,
              'laserFilterRange': laserFilterRange,
              'gaussianSmoothingSigma': gaussianSmoothingSigma,
}
save(configDict,fileNumString+'_configDict.dat')

# () Initialize the fitAllParallel structure
initializeFitAllParallel(p,prefix,maxTimesPerIndepParam=maxTimesPerIndepParam,
                         timeOrderSeed=timeSeed,numIndepParams=numInSampleTrials,
                         deltaNumDatapoints=deltaNumDatapoints)






